// Named timers
// by Jeremy Morton
// A jQuery plugin that uses setInterval to allow for easier timer and callback management, with named timer handles.

// For a nice jQuery plugin development pattern, see:
// http://www.learningjquery.com/2007/10/a-plugin-development-pattern

(function($) {
	// *****
	// Private functions
	// *****

	function isPureInt(input) {
		return typeof input === "number" && /^[0-9]+$/.test(input);
	};

	function isFunction(input) {
		return typeof input === "function";
	};

	function isString(input, nonEmpty) {
		return typeof input === "string" && (!nonEmpty || input !== "");
	};

	function isJquery(input) {
		return input instanceof jQuery;
	};

	function isUnspecified(input) {
		return input === null || input === undefined;
	};

	// *****
	// Class definitions
	// *****

	// *** NamedTimerInternal class ***
	function NamedTimerInternal(timerName) {
		var self = this;

		//
		// Init class instance
		//

		self._timerName = timerName;
		self._timerId = null;
		self._callbackFn = null;
		self._currentContext = null;
	}

	//
	// Public methods
	//

	// Gets this timer's name.
	NamedTimerInternal.prototype.getTimerName = function() {
		var self = this;

		return self._timerName;
	};

	// Starts this named timer with the given interval.
	// interval: The interval to wait between each tick (in milliseconds).
	// callImmediately: (optional) If true, calls the callback function immediately instead of waiting for first tick.
	NamedTimerInternal.prototype.startTimer = function(interval, callImmediately) {
		var self = this;

		if (self._timerId !== null) {
			throw "Timer '" + self._timerName + "' is already started.";
		}

		if (isUnspecified(interval)) { throw "Timer interval must be specified."; }
		if (!isPureInt(interval)) { throw "Timer interval must be an integer."; }
		if (self._callbackFn === null) { throw "No timer callback function has been specified."; }

		// Start the timer
		self._timerId = setInterval(function() { self._callbackFn(); }, interval);
		if (callImmediately) { self._callbackFn(); }
	};

	// Stops this named timer.  If the timer isn't currently running, does nothing.
	NamedTimerInternal.prototype.stopTimer = function() {
		var self = this;

		if (self._timerId !== null) {
			clearInterval(self._timerId);
			self._timerId = null;
		}
	};

	// Sets the callback function for this timer.
	// callbackFn: The callback function for this timer.
	NamedTimerInternal.prototype.setTimerFunction = function(callbackFn) {
		var self = this;

		self._callbackFn = callbackFn;
	};

	// Sets the current context in which the callback function is operating.
	// jqCtx: The context.
	NamedTimerInternal.prototype.setCurrentContext = function(jqCtx) {
		var self = this;

		self._currentContext = jqCtx;
	};

	// Gets the current context in which the callback function is operating.
	NamedTimerInternal.prototype.getCurrentContext = function() {
		var self = this;

		return self._currentContext;
	};

	// *** NamedTimer class - holds jQuery context and communicates with NamedTimerInternal ***
	function NamedTimer(internalTimer, jqCtx) {
		var self = this;

		//
		// Init class instance
		//

		self._internalTimer = internalTimer;
		self._jqCtx = null;

		if (!isUnspecified(jqCtx)) {
			if (!isJquery(jqCtx)) { throw "jqCtx must be a jQuery object."; }
			self._jqCtx = jqCtx;
		}
	}

	//
	// Public methods
	//

	// Gets this timer's name.
	NamedTimer.prototype.getTimerName = function() {
		var self = this;

		return self._internalTimer.getTimerName();
	};

	// Starts this named timer with the given interval.
	// interval: The interval to wait between each tick (in milliseconds).
	// callbackFn: (optional) If set, sets the callback function to run on each tick.
	// callImmediately: (optional) If true, calls the callback function immediately instead of waiting for first tick.
	NamedTimer.prototype.startTimer = function(interval, callbackFn, callImmediately) {
		var self = this;

		if (!isUnspecified(callbackFn)) {
			// Set timer function with appropriate jQuery context
			self.setTimerFunction(callbackFn);
		}

		self._internalTimer.startTimer(interval, callImmediately);
		return self;
	};

	// Stops this named timer.  If the timer isn't currently running, does nothing.
	NamedTimer.prototype.stopTimer = function() {
		var self = this;

		self._internalTimer.stopTimer();
		return self;
	};

	// Sets the callback function for this timer.
	// callbackFn: The callback function for this timer.
	NamedTimer.prototype.setTimerFunction = function(callbackFn) {
		var self = this;

		if (isUnspecified(callbackFn)) { throw "Timer callback function must be specified."; }
		if (!isFunction(callbackFn)) { throw "Timer callback function must be a function."; }

		if (self._jqCtx === null) {
			self._jqCtx = self._internalTimer.getCurrentContext();
		}
		else {
			self._internalTimer.setCurrentContext(self._jqCtx);
		}
		self._internalTimer.setTimerFunction(function() {
			if (self._jqCtx === null) { callbackFn(); }
			else { callbackFn.apply(self._jqCtx) }
		});
		return self;
	};

	// *** NamedTimerCollection class - holds collection of timers along with jQuery context, and allows for creation/deletion of timers ***
	function NamedTimerCollection(timers, jqCtx) {
		var self = this;

		//
		// Init class instance
		//

		self._timers = timers;
		self._jqCtx = null;

		if (!isUnspecified(jqCtx)) {
			if (!isJquery(jqCtx)) { throw "jqCtx must be a jQuery object."; }
			self._jqCtx = jqCtx;
		}
	}

	//
	// Public methods
	//

	// Create a new named timer.
	// timerName: The name of the new timer.
	NamedTimerCollection.prototype.createTimer = function(timerName) {
		var self = this;

		if (isUnspecified(timerName)) {
			throw "timerName must be specified.";
		}
		else if (timerName === "") {
			throw "timerName must not be an empty string.";
		}
		else if (timerName in self._timers) {
			throw "Timer '" + timerName + "' already exists.";
		}
		else {
			// Create internal timer with timerName and return its jQuery context wrapper
			return new NamedTimer(
				self._timers[timerName] = new NamedTimerInternal(timerName),
				self._jqCtx
			);
		}
	};

	NamedTimerCollection.prototype.deleteTimer = function(timerName) {
		var self = this;

		if (isUnspecified(timerName)) {
			throw "timerName must be specified.";
		}
		else if (timerName === "") {
			throw "timerName must not be an empty string.";
		}
		else if (!(timerName in self._timers)) {
			throw "Timer '" + timerName + "' doesn't exist.";
		}
		else {
			// Delete the specified timer, stopping it first if necessary
			self._timers[timerName].stopTimer();
			delete self._timers[timerName];
		}
	};

	// Get a list of existing named timers.
	NamedTimerCollection.prototype.getTimers = function() {
		var self = this;

		// Returns object with a jQuery context-wrapped timer for each internal timer
		var wrappedTimers = {};
		for (var timerIntKey in self._timers) {
			wrappedTimers[self._timers[timerIntKey].getTimerName()] = new NamedTimer(self._timers[timerIntKey], self._jqCtx);
		}

		return wrappedTimers;
	};

	// *****
	// Private plugin members
	// *****

	var priv = {};

	// Object to hold internal named timers collection
	priv.timers = {};

	// *****
	// Public plugin definitions
	// *****

	function mainEntryPoint(timerName, jqCtx) {
		if (isUnspecified(timerName)) {
			// Nothing specified, return named timer collection
			return new NamedTimerCollection(priv.timers, jqCtx);
		}
		else if (!isString(timerName, true)) {
			// Timer name specified is not a non-empty string
			throw "timerName must be a non-empty string.";
		}
		else if (priv.timers[timerName] === undefined) {
			// Timer specified, does not exist
			throw "Timer '" + timerName + "' does not exist.";
		}
		else {
			// Existing timer specified, return it
			return new NamedTimer(priv.timers[timerName], jqCtx);
		}
	};

	// Register as AMD or in-browser jQuery
	if ($ == null) {
		module.exports = function(timerName) {
			return mainEntryPoint(timerName);
		};
	}
	else {
		// Function to be called directly against jQuery, eg. jQuery.namedTimers()
		$.namedTimers = mainEntryPoint;

		// Function that can be used with jQuery accessors, eg. jQuery('.myClass').namedTimers()
		$.fn.namedTimers = function(timerName) {
			// $.namedTimers()         // --> NamedTimerCollection / NamedTimer
			//   .createTimer("foo")   // NamedTimerCollection
			//   .setTimerFunction(... // NamedTimer (jqCtx: null) -> NamedTimerInternal

			// $.namedTimers("foo")    // --> NamedTimerCollection / NamedTimer
			//   .setTimerFunction(... // NamedTimer (jqCtx: null) -> NamedTimerInternal

			// $('#divXyz')            // jQuery
			//   .namedTimers()        // --> NamedTimerCollection / NamedTimer
			//   .createTimer("foo")   // NamedTimerCollection
			//   .setTimerFunction(... // NamedTimer (jqCtx: '#divXyz') -> NamedTimerInternal

			// var abc = $('#divXyz')  // jQuery
			//   .namedTimers()        // --> NamedTimerCollection / NamedTimer
			//   .createTimer("foo");  // NamedTimerCollection
			// abc                     // NamedTimer (jqCtx: '#divXyz') -> NamedTimerInternal
			//   .setTimerFunction(... // NamedTimer (jqCtx: '#divXyz') -> NamedTimerInternal

			// var abc = $('#divXyz')  // jQuery
			//   .namedTimers()        // --> NamedTimerCollection / NamedTimer
			//   ;                     // NamedTimerCollection

			// $('#divXyz')            // jQuery
			//   .namedTimers("foo")   // --> NamedTimerCollection / NamedTimer
			//   .setTimerFunction(... // NamedTimer (jqCtx: '#divXyz') -> NamedTimerInternal

			// $('#divXyz')            // jQuery
			//   .namedTimers()        // --> NamedTimerCollection / NamedTimer
			//   .getTimers()          // Array[] of NamedTimer (jqCtx: [varied])
			//   [2]                   // NamedTimer (jqCtx: '#divXyz') -> NamedTimerInternal
			//   .setTimerFunction(... // NamedTimer (jqCtx: '#divXyz') -> NamedTimerInternal

			return $.namedTimers(timerName, this); // 'this', at this point in time, is jqCtx (jQuery context)
		};
	}
})(
	(function(){
		if (typeof module === "object" && typeof module.exports === "object") {
			return null;
		}
		else if (typeof jQuery !== "undefined") {
			return jQuery;
		}
		else {
			throw "No AMD export possible and jQuery isn't defined!  Unable to initialize.";
		}
	})()
);
